/*
    futebus - FUse TExt BUS - fuse based IPC bus for text lines
    Copyright (C) 2023  Tibor 'Igor2' Palinkas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    Project page: http://www.repo.hu/projects/futebus
    VCS: svn://svn.repo.hu/futebus/trunk
    Contact: see doc/contact.html
*/

#include "futebus.h"
#include "message.h"
#include "channel.h"
#include "dispatch.h"

#include "client.h"

/* Announce join/leave of clients to the master on an 'S' channel */
static void master_announce(channel_t *chan, uint32_t client_id, char op)
{
	message_t *msg;
	char tmp[256];
	int len;

	if ((chan->type != 'S') || (chan->master == NULL))
		return;

	len = sprintf(tmp, "{%lx:%c}\n", (unsigned long)client_id, op);

	msg = message_alloc(chan, tmp, len);
	msg->type = MT_CTRL; /* do not prefix with addrss */
	insert_msg_in_rlist(chan->master, msg, 0, 0, 1);
}

client_t *client_alloc_to_chan(channel_t *chan, int rd, int wr, int as_master)
{
	client_t *client = calloc(sizeof(client_t), 1);

	if (client == NULL)
		return NULL;

	mutex_lock(&chan->lock);

	if (as_master && (chan->master != NULL)) {
		mutex_unlock(&chan->lock);
		return NULL;
	}

	client->id = chan->next_client_id++;
	client->rd = rd;
	client->wr = wr;
	client->master = as_master;
	mutex_init(&client->poll_lock);
	mutex_init(&client->rlist_lock);
	notify_init(&client->rlist_notify);

	client->chan = chan;
	gdl_append(&chan->clients, client, link);
	if (as_master)
		chan->master = client;

	logprintf("NEW CLIENT: %p %d rd=%d wr=%d\n", chan, client->id, client->rd, client->wr);

	mutex_unlock(&chan->lock);

	if (!as_master)
		master_announce(chan, client->id, 'o');

	return client;
}

void client_free(client_t *client)
{
	int n, destroyed = 0, waiters;

	client->dead = 1;

	/* do not destroy if others are waiting for this client's rlist, until they quit */
	mutex_lock(&client->rlist_lock);
	waiters = client->rlist_waiters > 0;
	mutex_unlock(&client->rlist_lock);
	if (waiters)
		return;

	if (client->freeing)
		return;

	client->freeing = 1;

	if (client->chan->master != client)
		master_announce(client->chan, client->id, 'c');

	mutex_lock(&client->chan->lock);
	gdl_remove(&client->chan->clients, client, link);
	if (client->chan->master == client) {
		client->chan->master = NULL;
		destroyed = chan_unlink(client->chan);
	}
	if (!destroyed) {
		mutex_unlock(&client->chan->lock);
		if ((client->chan != statch) && client->chan->zombie && (client->chan->clients.length == 0)) {
			chan_destroy(client->chan);
			destroyed = 1;
		}
	}

	mutex_lock(&client->rlist_lock);
	gds_uninit(&client->rbuf);
	gds_uninit(&client->wbuf);
	if (!destroyed) {
		/* if channel got destroyed messages are all free'd already */
		for(n = 0; n < client->rlist.used; n++)
			message_unref(client->rlist.array[n]);
	}
	vtp0_uninit(&client->rlist);
	mutex_unlock(&client->rlist_lock);

	futebus_poll_uninit(client);
	mutex_uninit(&client->poll_lock);
	mutex_uninit(&client->rlist_lock);
	notify_uninit(&client->rlist_notify);

	free(client);
}

void client_waiter_inc(client_t *client)
{
	assert(!client->dead);
	mutex_lock(&client->rlist_lock);
	client->rlist_waiters++;
	mutex_unlock(&client->rlist_lock);
}

void client_waiter_dec(client_t *client)
{
	int waiters;

	mutex_lock(&client->rlist_lock);
	assert(client->rlist_waiters != 0);
	client->rlist_waiters--;
	waiters = client->rlist_waiters > 0;
	mutex_unlock(&client->rlist_lock);

	/* there is no race here: nobody can call client_waiter_inc() on a dead client */

	if (client->dead && !waiters)
		client_free(client);
}
